;******************************************************************************
;                                                                             *
;    Filename: Pendulo.asm                                                    *
;    Date: 12/02/2010                                                         *
;    File Version: 1.0                                                       *
;                                                                             *
;    Author:  Daniel Hctor Stolfi Rosso                                      *
;    Company: www.danielstolfi.com                                            *
;                                                                             * 
;******************************************************************************
;                                                                             *
;    Files required: P18F4520.INC                                             *
;                                                                             *
;******************************************************************************

	LIST P=18F4520, F=INHX32 ;directive to define processor and file format
	#include <P18F4520.INC>	 ;processor specific variable definitions

;******************************************************************************
;Configuration bits

	CONFIG	OSC = INTIO67	; Internal oscillator block, port function on RA6 and RA7
	CONFIG	WDT = OFF		; WDT off

;******************************************************************************
; Constants

_DIR_STOP	EQU 0x00	; Stop head
_DIR_LEFT	EQU 0x01	; Move head to the left
_DIR_RIGHT	EQU 0x02	; Move head to the right

;SEQ_1		EQU B'00000001'	; 1 = 12
;SEQ_2		EQU B'00001000'	; 2 = 43
;SEQ_3		EQU B'00000010'	; 3 = 21
;SEQ_4		EQU B'00000100'	; 4 = 34
;
_SEQ_1		EQU B'00000101'	; 1 = 12-34
_SEQ_2		EQU B'00001001'	; 2 = 12-43
_SEQ_3		EQU B'00001010'	; 3 = 21-43
_SEQ_4		EQU B'00000110'	; 4 = 21-34


;******************************************************************************
;Variable definitions
; These variables are only needed if low priority interrupts are used. 
; More variables may be needed to store other special function registers used
; in the interrupt routines.

		UDATA_ACS

DIRECTION	RES 1	; _DIR_STOP, _DIR_LEFT or _DIR_RIGHT
SEQ			RES 1	; 1 - 4
DELAY		RES 1
ACTION		RES 1
COUNT		RES 1
LOOP		RES 1	; Looping var
ERR			RES 1	; e(k)
ERR_1		RES 1	; e(k-1)
ERR_DIF		RES 1	; e(k) - e(k-1)
ENDS		RES 1	; ENDS<0> Left, ENDS<1> Right
ADL			RES 1
ADH			RES 1
ANGLE		RES 1
HEAD		RES 1
TEMP		RES 1
TEL_ACT		RES 1

;******************************************************************************
;EEPROM data
; Data to be programmed into the Data EEPROM is defined here

DATA_EEPROM	CODE	0xf00000

;******************************************************************************
;Reset vector
; This code will start executing when a reset occurs.

RESET_VECTOR	CODE	0x0000

		goto	Main		;go to start of main code

;******************************************************************************
;High priority interrupt vector

HI_INT_VECTOR	CODE	0x0008

		bra	HighInt		;go to high priority interrupt routine

;******************************************************************************
;Low priority interrupt vector

LOW_INT_VECTOR	CODE	0x0018

		bra	LowInt		;go to low priority interrupt routine

;******************************************************************************
;High priority interrupt routine
; The high priority interrupt code is placed here.

		CODE

HighInt:
		;;;;;;;;;;;;;;;;;;;;;;;;; each 1.54 mS @ 8 MHz
		movlw	.64
		movwf	TMR0L			; Adjust of timer
		
		movf	ANGLE,W			; Signed Byte [-128 - +127]
		call	Tx				; Angle -> RS232
		
		;;; Test Ends
		btfss	PORTA,RA6		; if left end is reached (RA6=0)
		bsf		ENDS,0			; turn on ENDS<0>
		btfss	PORTA,RA7		; if right end is reached (RA7=0)
		bsf		ENDS,1			; turn on ENDS<1>		
		btfsc	ENDS,0			; if Left End
		goto	Move0			; don't move head
		btfsc	ENDS,1			; if Right End
		goto	Move0			; don't move head
		
		btfss	PORTA,RA1		; if calibrating (RA1=0)
		goto	Move0			; don't move head
								; else calculate movement
		
		;;; Calculate ERR_DIF
		movf	ERR_1,W
		subwf	ERR,W
		movwf	ERR_DIF			; ERR(k) - ERR(k-1) -> ERR_DIF
		;;; Save current error
		movff	ERR,ERR_1		; ERR(k) -> ERR(k-1)
				
Control
		;;;;;;;;;;;;;;;;;;;;;;;;; Proportional Control	
		movlw	B'00001100'		; Mask b2 and b3
		andwf	PORTA,W			; KP
		rrncf	WREG
		rrncf	WREG
		incf	WREG			; KP (1,2,3 or 4)
		
		;;;;;;;;;;;;;;;;;;;;;;;;; Action Proportional
		mulwf	ERR				; KP * ERR(k) -> PRODH:PRODL
		movff	PRODL,ACTION	; Action P in ACTION [-16 - +16]

		;;;;;;;;;;;;;;;;;;;;;;;;; Derivative Control
		movlw	B'00110000'		; Mask b4 and b5
		andwf	PORTA,W			; KD
		rrncf	WREG
		rrncf	WREG
		rrncf	WREG
		rrncf	WREG			; KD (0,1,2 or 3)		
	
		;;;;;;;;;;;;;;;;;;;;;;;;; Action Derivative
		mulwf	ERR_DIF			; KD * [ERR(k) - ERR(k-1)] -> PRODH:PRODL
		btfsc	ERR_DIF,7
		subwf	PRODH
		movf	PRODL,W
		rlcf	PRODH			; sign PRODH:PRODL -> carry
		rrcf	WREG			; Divides by 2 (KD = 0, 0.5, 1, 1.5)
		addwf	ACTION
		
		movff	ACTION,TEL_ACT	; Telemetry

		;;;;;;;;;;;;;;;;;;;;;;;;; Resolve direction

		movlw	0x00
		cpfseq	ACTION			; if ACTION <> 0
		goto	Direction		; goto Direction: move head
		goto	Move0			; else No Error: Don't move head

Direction
		movlw	_DIR_RIGHT
		movwf	DIRECTION		; Move to the right by default
		btfss	ACTION,7		; if ACTION >= 0
		goto	Quantizer		; goto Quantizer
		
		;;; Positivate
		negf	ACTION			; -ACTION -> ACTION
		movlw	_DIR_LEFT
		movwf	DIRECTION		; _DIR_LEFT -> DIRECTION
		
Quantizer
		;;;;;;;;;;;;;;;;;;;;;;;;; Quantize in 4 levels
		movff	ACTION,DELAY	
		bsf		STATUS,C		; Set Carry
		movlw	.5
		subfwb	DELAY			; 5 - DELAY -> DELAY
		btfsc	DELAY,7			; if DELAY < 0
		clrf	DELAY			; 0 -> DELAY
 		movlw	.0
		cpfseq	DELAY			
		goto	Dir				; if DELAY = 0
		movlw	.1
		movwf	DELAY			; 1 -> DELAY
		
Dir
		incf	COUNT			; Inc COUNT
		movf 	DELAY,W
		cpfsgt	COUNT			; if COUNT <= DELAY
		goto	Move0			; do nothing (Stop & Wait)
		movlw	0x01
		movwf	COUNT			; else begin count again
								; Resolves direction of movement
		movlw	_DIR_LEFT
		cpfseq	DIRECTION		; if DIRECTION equals DIR_LEFT
		goto	Right			; jump to Right
		goto 	Left			; else jump to Left

Left							; Moves to the Left
		movlw	'L'
		movwf	HEAD			; Telemetry: Move Head to the Left

		incf	SEQ				; Inc SEQ
		movlw	0x04
		cpfsgt	SEQ				; if SEQ <= 4
		goto	Move			; jump to Move
		movlw	0x01		
		movwf	SEQ				; else SEQ = 1
		goto	Move			; jump to Move

Right							; Moves to the Right
		movlw	'R'
		movwf	HEAD			; Telemetry: Move Head to the Right
		
		decfsz	SEQ				; Dec SEQ, if SEQ <> 0
		goto	Move			; Jump to Move
		movlw	0x04
		movwf	SEQ				; else SEQ = 4

Move							; moves head		
		movlw	0x01
		cpfseq	SEQ				; if SEQ <> 1
		goto	Move2			; go to test for _SEQ_2
		movlw	_SEQ_1			; else w = _SEQ_1
		goto	EndMove		
Move2
		movlw	0x02
		cpfseq	SEQ				; if SEQ <> 2
		goto	Move3			; go to test for _SEQ_3
		movlw	_SEQ_2			; else w = _SEQ_2
		goto	EndMove
Move3
		movlw	0x03
		cpfseq	SEQ				; if SEQ <> 3
		goto	Move4			; go to test for _SEQ_3
		movlw	_SEQ_3			; else w = _SEQ_3
		goto	EndMove
Move4
		movlw	_SEQ_4			; w = _SEQ_4
		goto	EndMove
		
Move0
		movlw	'S'
		movwf	HEAD			; Telemetry: Stop Head

		movlw	0x00			; remove voltage of motor

EndMove
		movwf	PORTB			; puts W in PORTB
		
EndHighInt
		movf	TEL_ACT,W
		call 	Tx				; Signed Action -> RS232
		movf	HEAD,W
		call	Tx				; Head -> RS232

		bcf		INTCON,TMR0IF	; clear flag timer int.
		retfie	FAST			; return form interrupt

;******************************************************************************
;Low priority interrupt routine
; The low priority interrupt code is placed here.
; This code can be removed if low priority interrupts are not used.

LowInt:
	
		retfie

;******************************************************************************
;Start of main program

Main:

Init
	; Clock
	bsf		OSCCON,IRCF0
	bsf		OSCCON,IRCF1
	bsf		OSCCON,IRCF2	; 8 MHz
	bsf		OSCCON,SCS1		; Internal Oscilator

	; Initialize PORT B: Drive the motor
	clrf 	PORTB
	clrf 	LATB
	clrf	TRISB			; all outputs

	; Initialize PORT E: Calibrating LED
	clrf 	PORTE
	clrf 	LATE
	clrf	TRISE			; all outputs

	; Initialize PORT A: Error, Cal, Kp & Kd
	clrf 	PORTA
	clrf 	LATA
	setf	TRISA			; all inputs
	
	; Initialize EUSART (Port C): Telemetry
	setf	TRISC			; all inputs
	
	bcf		TXSTA,TX9		; 8-bit transmission
	bcf		TXSTA,SYNC		; Asynchronous mode
	bsf		TXSTA,BRGH		; High speed
	
	bsf		BAUDCON,BRG16	; 8-bit Baud Rate Generator
	
	movlw	.34				; Internal counter
	movwf	SPBRG			; 57600 Bauds
	
	bsf		RCSTA,SPEN		; Enable EUSART	
	bsf		TXSTA,TXEN		; Enable the transmission
		
	; Initialize ADC
	movlw	b'00000000'
	movwf	ADCON0			; select channel 0 (AN0)
	
	movlw	b'00001110'
	movwf	ADCON1			; Only input AN0 is analog
	
	movlw	b'10000110'		; Read Right justified
	movwf	ADCON2			; Clock: FOSC/64
			
	bsf		ADCON0,ADON		; converter enabled
		
	; Initialize Comparator: Turn off
	movlw	0x07
	movwf	CMCON			; comparator disabled		
		
	; Initialize timer0: Control cycle							
							; 8 bits mode
	movlw	b'01010011'		; Internal clock
	movwf	T0CON			; prescaler (divides by 16)
		
	; Interruption
	bcf		RCON,IPEN		; disable priority levels
	bsf		INTCON,GIE		; enable all unmasked interrupts
	bcf		INTCON,PEIE		; disable all unmasked peripheral interrupts
	bcf		INTCON,TMR0IF	; clear flag timer int.
	bsf		INTCON,TMR0IE	; activate timer overflow interruption
		
	; Initialize vars
	movlw	_SEQ_1			; Motor sequence Literal
	movwf	SEQ				; Initialize Sequence
	movlw	_DIR_STOP		; Direction Literal
	movwf	DIRECTION		; Stopped
	clrf	ACTION			; Controler Action
	movlw	0xFF			
	movwf	DELAY			; Initial Delay
	movlw	0x01
	movwf	COUNT			; Initial Count
	clrf	ENDS			; not in left end, not in right end
	clrf	ERR_1			; Initial E(k-1)
	clrf	ERR				; Initial E(k)
	
	;;;;;;;;;;;;;;;;;;;;;;;;; Telemetry
	clrf	TEL_ACT			; Control
	movlw	'S'
	movwf	HEAD			; Head stopped
	clrf	ANGLE			; Initial angle
	;;;;;;;;;;;;;;;;;;;;;;;;;
	
;**********************************************************
; Here we go!
;**********************************************************
	bsf		T0CON,TMR0ON	; activate timer
	
Loop
	bsf		ADCON0,NOT_DONE	; Begins ADC conversion

Read
	btfsc	ADCON0,NOT_DONE
	goto 	Read			; Polling for ADC value

	movff	ADRESH,ADH		; ADRESH -> ADH
	movff	ADRESL,ADL		; ADRESL -> ADL

	call	Calc_E			; Convert ADC value in SPEED and DIRECTION
							; returns error in WREG [-4 - 4]							
	movwf	ERR				; Updates ERR

	movlw	0x02
	movwf	LOOP			; Counter
	
	movlw	0x00
	bcf		PORTE,RE0		; Turn off Center LED (calibrating)
	cpfseq	ERR
	goto	ADCDelay		; if ERR = 0
	bsf		PORTE,RE0		; Turn on Center LED (calibrating)
	
ADCDelay
	decfsz	LOOP
	goto	ADCDelay		; Wait a minimum of 2 TADs before another conversion
	
	goto 	Loop

;**********************************************************
; Transform Error
;**********************************************************
Calc_E
	;;;;;;;;;;;;;;;;;;;;;;;;; Range test (ADH value)
	movlw	B'00000011'		; Mask b0 & b1
	andwf	ADH
	movlw	.3
	cpfslt	ADH				; if ERROR > 1536 (11 00000000)
	goto	OutGreater		; Out of Range (Greater)
	movlw	.0
	cpfsgt	ADH				; if ERROR < 512 (01 00000000)
	goto	OutLower		; Out of Range (Lower)
	movlw	.1
	cpfseq	ADH				; if ERROR is positive (0x0200 - 0x02FF)
	goto	Greater 		; move to the right (greater)
							; else move to the left (lower) 
							; (0x01FF - 0x0100)

Lower
	;;;;;;;;;;;;;;;;;;;;;;;;; Telemetry
	movff	ADL,ANGLE
	movlw	.128
	cpfsgt	ANGLE			; if ANGLE < -128
	movwf	ANGLE			; ANGLE = -128
	;;;;;;;;;;;;;;;;;;;;;;;;;
	movlw	.252			; Saturation Check (-4)
	cpfsgt	ADL				; if ADL <= -4
	retlw	.252			; Saturation: return -4
	movf	ADL,W			; else
	return					; return Error
	
Greater
	;;;;;;;;;;;;;;;;;;;;;;;;; Telemetry
	movff	ADL,ANGLE
	movlw	.127
	cpfslt	ANGLE			; if ANGLE > 127
	movwf	ANGLE			; ANGLE = 127
	;;;;;;;;;;;;;;;;;;;;;;;;;
	movlw	.4				; Saturation Check
	cpfslt	ADL				; if ADL >= 4
	retlw	.4				; Saturation: return 4
	movf	ADL,W			; else
	return					; return error

OutGreater
	;;;;;;;;;;;;;;;;;;;;;;;;; Telemetry
	movlw	.127
	movwf	ANGLE			; ANGLE = 127
	;;;;;;;;;;;;;;;;;;;;;;;;;	
	retlw	.4				; Saturation: return 4
	
OutLower
	;;;;;;;;;;;;;;;;;;;;;;;;; Telemetry
	movlw	.128
	movwf	ANGLE			; ANGLE = -128
	;;;;;;;;;;;;;;;;;;;;;;;;;	
	retlw	.252			; Saturation: return -4

;**********************************************************
Tx							; RS232
	btfss	TXSTA,TRMT		; 1 = TSR Empty
	goto	Tx				; Wait to transmit
	movwf	TXREG			; Transmits WREG
	return
	
;End of program

		END
